/*==============================================================================
  File Name: bitStacker.cpp
  Author   : ccortiz
  
  Description:
  Class for stack and unstack bits in a byte buffer.
  ==============================================================================
*/

#include "bitStack.h"
#include <stdio.h>

#if  _ALLOW_MALLOC_ == 0
/* Reserve maximum amount of memory allowed for BITSTACK. */
struct c_bitStack_Privates bitStack_Instances[_MAX_BITSTACK_INSTANCES_];
#else
#include <stdlib.h>
#endif

/* Private functions. */
static void swap16(uint16 *data);
static void swap32(uint32 *data);
static enum endianness getSystemEndianness(void);
static void resetUnstack(uint16 bitStackDescriptor);

enum endianness endianType; /*Indica el endianness del sistema.*/
bitStack_PrivatesPtr bitStack_Objs[_MAX_BITSTACK_INSTANCES_];
uint16 nBitStackObjs = 0;	/*Indica el numero de instancias realizadas de la clase bitStack.*/

int new_bitStack(void){
	bitStack_PrivatesPtr newObj;
	
	endianType = getSystemEndianness();

	if (nBitStackObjs>_MAX_BITSTACK_INSTANCES_ -1){
		/* No more instances allowed. */
		return -1;
	}

#if _ALLOW_MALLOC_==0
	newObj =&bitStack_Instances[nObjs];
#else
	newObj = (bitStack_PrivatesPtr)malloc(sizeof(struct c_bitStack_Privates));
#endif

	if (newObj==0){
		return -1;
	}
	else{
		/* Inicializa  variables del objeto. */

		bitStack_Objs[nBitStackObjs] = newObj;
	}

	nBitStackObjs++;

	/* Configure the default endianness for outputs. */
	newObj->outputEndianness =_BIG_ENDIAN_;

	clearStack(nBitStackObjs-1);
	newObj->unstackIndex=0;

	return nBitStackObjs-1;
}

void clearStack(uint16 bitStackDescriptor){
	uint32 i = 0;

	bitStack_Objs[bitStackDescriptor]->auxUsedBits = 0;
	bitStack_Objs[bitStackDescriptor]->stackSize   = 0;

	for (i=0; i<_MAX_DATA_STACK_SIZE_; i++){
		bitStack_Objs[bitStackDescriptor]->dataStack[i] = 0;
	}

	/* Reset unstack variables. */
	resetUnstack(bitStackDescriptor);
}

void resetUnstack(uint16 bitStackDescriptor){
	bitStack_Objs[bitStackDescriptor]->firstUnstak = _TRUE_;
}

enum endianness getSystemEndianness(void){
	sint32 i = 1;
	char *p = (char *) &i;

	if (p[0] == 1) {
		return _LITTLE_ENDIAN_;
	} else {
		return _BIG_ENDIAN_;
	}
}

sint32 stackData(uint16 bitStackDescriptor, unsigned char* data, uint32 dataLength){

	char dataBuffer[8];

	byte bitState;
	byte dataByte;
	byte auxByte;
	uint32 dataUsedBits = 0;
	uint32 byteCounter=0;
	uint32 i=0;

	memcpy(dataBuffer, data, (dataLength/8) + (((dataLength%8)>0)?1:0) );
	
	if (dataLength==16 || dataLength==32){
		if (endianType != bitStack_Objs[bitStackDescriptor]->outputEndianness){
			if (dataLength==32){
				swap32((uint32*)dataBuffer);
			}
			else{
				swap16((uint16*)dataBuffer);
			}
		}
	}
	else{

		if (endianType==_BIG_ENDIAN_){
			if (dataLength>16 && dataLength<32){
				swap32((uint32*)dataBuffer);
			}
			else{
				if (dataLength>8 && dataLength<16){
					swap16((uint16*)dataBuffer);
				}
				else{
					/*Aqui habria que contemplar el caso de campos de entre 32 y 64 bits*/
				}
			}
		}
	}


	dataByte = *dataBuffer;
	auxByte  = bitStack_Objs[bitStackDescriptor]->dataStack[bitStack_Objs[bitStackDescriptor]->stackSize];

	for (i=0; i<dataLength; i++){
		bitState = dataByte & 0x1;
		auxByte = auxByte | (bitState<<bitStack_Objs[bitStackDescriptor]->auxUsedBits);
		bitStack_Objs[bitStackDescriptor]->auxUsedBits++;
		dataUsedBits++;

		if (bitStack_Objs[bitStackDescriptor]->auxUsedBits==8){
			bitStack_Objs[bitStackDescriptor]->auxUsedBits = 0;
			bitStack_Objs[bitStackDescriptor]->dataStack[bitStack_Objs[bitStackDescriptor]->stackSize++] = auxByte;
			auxByte = 0;
		}

		if (dataUsedBits==8 && (i+1)<dataLength)
		{
			dataUsedBits = 0;
			byteCounter++;
			dataByte = *(dataBuffer + byteCounter);
		}
		else{
			dataByte=dataByte>>1;
		}
	}

	if (bitStack_Objs[bitStackDescriptor]->auxUsedBits!=0){
		bitStack_Objs[bitStackDescriptor]->dataStack[bitStack_Objs[bitStackDescriptor]->stackSize] = auxByte;
	}

	return 0;
}


uint32 unstackData(uint16 bitStackDescriptor, uint8 nBits){

	uint32 aux=0;
	uint32 i=0;
	uint32 j=0;

	if (bitStack_Objs[bitStackDescriptor]->firstUnstak){
		bitStack_Objs[bitStackDescriptor]->firstUnstak= _FALSE_;
		bitStack_Objs[bitStackDescriptor]->unstackedBit=0;
		bitStack_Objs[bitStackDescriptor]->unstackedByte = bitStack_Objs[bitStackDescriptor]->dataStack[0];
		bitStack_Objs[bitStackDescriptor]->unstackIndex=0;
	}

	if (nBits>32){
		_DBGMSG_("ERROR! Can't unstack more than 64bit at time.");
		return 0;
	}

	if (bitStack_Objs[bitStackDescriptor]->unstackIndex >= getStackSize(bitStackDescriptor)){
		_DBGMSG_("ERROR! Empty Buffer!");
		return 0;
	}

	for (i=0; i<nBits; i++){
		aux = aux >>1;

		if (bitStack_Objs[bitStackDescriptor]->unstackedByte & 0x1){
			aux = aux | 0x80000000L;
		}

		bitStack_Objs[bitStackDescriptor]->unstackedByte = bitStack_Objs[bitStackDescriptor]->unstackedByte >> 1;

		bitStack_Objs[bitStackDescriptor]->unstackedBit++;
		if (bitStack_Objs[bitStackDescriptor]->unstackedBit >= 8){
			bitStack_Objs[bitStackDescriptor]->unstackedBit=0;
			if (bitStack_Objs[bitStackDescriptor]->unstackIndex<getStackSize(bitStackDescriptor)){
				bitStack_Objs[bitStackDescriptor]->unstackIndex++;
				bitStack_Objs[bitStackDescriptor]->unstackedByte = bitStack_Objs[bitStackDescriptor]->dataStack[bitStack_Objs[bitStackDescriptor]->unstackIndex];
			}
			else{
				bitStack_Objs[bitStackDescriptor]->unstackedByte=0;
			}
		}
	}

	for (j=0; j<(32-nBits); j++){
		aux = aux >> 1;
	}

	/* Make swappings. */
	if (bitStack_Objs[bitStackDescriptor]->outputEndianness == _BIG_ENDIAN_){
		if (nBits==16){
			uint16 data = aux;
			swap16(&data);
			aux = data;
		}

		if (nBits==32){
			uint32 data = aux;
			swap32(&data);
			aux = data;
		}
	}
	return aux;
}

void showStack(uint16 bitStackDescriptor){

	uint32 i=0;

	for (i=0; i<getStackSize(bitStackDescriptor); i++){
		printf("%.2X ",(uint32)(bitStack_Objs[bitStackDescriptor]->dataStack[i]));
	}
}

uint32 getStackSize(uint16 bitStackDescriptor){

	uint32 showLength = bitStack_Objs[bitStackDescriptor]->stackSize;

	if (bitStack_Objs[bitStackDescriptor]->auxUsedBits>0){
		showLength++;
	}

	return showLength;
}

void setOutputEndianness(uint16 bitStackDescriptor, enum endianness endianType){
	bitStack_Objs[bitStackDescriptor]->outputEndianness = endianType;
}
	
byte_ptr getDataBuffer(uint16 bitStackDescriptor){
	return bitStack_Objs[bitStackDescriptor]->dataStack;
}

sint32 setDataBuffer(uint16 bitStackDescriptor, byte_ptr data,uint16 dataLength){
	if (dataLength<_MAX_DATA_STACK_SIZE_){
		memcpy(bitStack_Objs[bitStackDescriptor]->dataStack,data,dataLength);
		bitStack_Objs[bitStackDescriptor]->stackSize = dataLength;
		return 0;
	}

	return -1;
}

void swap16(uint16 *data){
	byte b0 = ((*data)&0x00FF)>>0;
	byte b1 = ((*data)&0xFF00)>>8;
	*data = (b1) | (b0<<8);
}

void swap32(uint32 *data){
	byte b0 = ((*data)&0x000000FF)>>0;
	byte b1 = ((*data)&0x0000FF00)>>8;
	byte b2 = ((*data)&0x00FF0000)>>16;
	byte b3 = ((*data)&0xFF000000)>>24;
	*data = b3 | (b2<<8) | (b1<<16) | (b0<<24);
}

